#!/usr/bin/env python2
# coding:utf-8
'''
Summary runtime infomation

Usage: python runtime.py <pairfile> <logdir> <outfile>
'''


import os
import sys
import re

from datetime import datetime


def listdir(path):
    for a, b, c in os.walk(path):
        for i in c:
            yield os.path.join(a, i)


def getSample(pairfile):
    samples = []
    with open(pairfile) as fi:
        for line in fi:
            if not line.strip() or line.startswith("#"):
                continue
            line = line.strip().split()
            samples.append(line[0])
    return samples


def sumtime(pairfile, logdir, timefile):
    logfiles = listdir(logdir)
    samples = getSample(pairfile)
    sampletimes = {}
    timeinfo = {}
    for f in logfiles:
        filename = os.path.basename(f)
        for s in samples:
            if s in filename:
                break
        else:
            continue
        with open(f) as fi:
            ctx = fi.read()
        intimes = re.findall(
            "\n\[(\d{4}-\d{2}-\d{2} \d{2}:\d{2}:\d{2})\]", ctx)
        dtime = sorted([datetime.strptime(i, "%Y-%m-%d %X") for i in intimes])
        assert len(dtime) >= 3
        sampletimes.setdefault(s, {}).setdefault("start", []).append(dtime[0])
        sampletimes.setdefault(s, {}).setdefault("end", []).append(dtime[-1])

    for sn, t in sampletimes.items():
        s = sorted(t["start"])[0]
        e = sorted(t["end"])[-1]
        timeinfo[sn] = (s, e)

    with open(timefile, "w") as fo:
        fo.write("Sample\tstart_time\tend_time\ttime_used(hour)\n")
        for sn, t in timeinfo.items():
            s = t[1] - t[0]
            fo.write(sn + "\t" + t[0].strftime("%Y/%m/%d,%X") + "\t" + t[1].strftime(
                "%Y/%m/%d,%X") + "\t" + str(round(s.total_seconds()/3600, 2)) + "\n")


if __name__ == "__main__":
    if len(sys.argv) != 4 or "-h" in sys.argv or "--help" in sys.argv:
        print __doc__
        sys.exit(1)
    pairfile, logdir, timefile = sys.argv[1:][:3]
    sumtime(pairfile, logdir, timefile)



#!/usr/bin/env python2
# coding:utf-8

import os
import sys
import time
import signal
import argparse
import logging

from subprocess import call, PIPE
from threading import Thread
from datetime import datetime
from collections import Counter

from qsub import qsub
from version import __version__


class ParseSingal(Thread):
    def __init__(self):
        super(ParseSingal, self).__init__()
        signal.signal(signal.SIGINT, self.signal_handler)
        signal.signal(signal.SIGTERM, self.signal_handler)

    def run(self):
        time.sleep(1)

    def signal_handler(self, signum, frame):
        if clear:
            call('qdel "*_%d"' % os.getpid(),
                 shell=True, stderr=PIPE, stdout=PIPE)
            pid = os.getpid()
            gid = os.getpgid(pid)
            call("kill -9 -%d" % gid, shell=True, stderr=PIPE, stdout=PIPE)
        sumJobs(qjobs)
        sys.exit(signum)


def parseArgs():
    parser = argparse.ArgumentParser(
        description="For manger submit your jobs in a job file.")
    parser.add_argument("-n", "--num", type=int,
                        help="the max job number runing at the same time, default: 1000", default=1000, metavar="<int>")
    parser.add_argument("-j", "--jobfile", type=str, required=True,
                        help="the input jobfile", metavar="<jobfile>")
    parser.add_argument('-v', '--version',
                        action='version', version="%(prog)s v" + __version__)
    parser.add_argument('-i', '--injname', help="job names you need to run, default: all jobnames of you job file",
                        nargs="*", type=str, metavar="<str>")
    parser.add_argument('-s', '--start', help="job beginning with the number you given, 1 by default",
                        type=int, default=1, metavar="<int>")
    parser.add_argument('-e', '--end', help="job ending with the number you given, last job by default",
                        type=int, metavar="<int>")
    parser.add_argument('-r', '--resub', help="rebsub you job when error, 0 or minus means do not re-submit, 3 times by default",
                        type=int, default=3, metavar="<int>")
    parser.add_argument('-ivs', '--resubivs', help="rebsub interval seconds, 2 by default",
                        type=int, default=2, metavar="<int>")
    parser.add_argument("-m", '--mode', type=str, default="sge", choices=[
                        "sge", "localhost"], help="the mode to submit your jobs, 'sge' by default.")
    parser.add_argument("-nc", '--noclean', action="store_false", help="whether to clean all jobs or subprocess created by this programe when the main process exits, default: clean.",
                        default=True)
    return parser.parse_args()


def Mylog(logfile=None, level="info"):
    logger = logging.getLogger()
    f = logging.Formatter(
        '[%(levelname)s %(processName)s %(asctime)s %(funcName)s] %(message)s')
    if logfile is None:
        h = logging.StreamHandler(sys.stdout)  # default: sys.stderr
    else:
        h = logging.FileHandler(logfile, mode='a')
    h.setFormatter(f)
    logger.addHandler(h)
    if level.lower() == "info":
        logger.setLevel(logging.INFO)
    elif level.lower() == "debug":
        logger.setLevel(logging.DEBUG)
    return logger


def sumJobs(qjobs):
    thisrunjobs = set([j.name for j in qjobs.jobs])
    realrunjobs = thisrunjobs - qjobs.has_success
    realrunsuccess = qjobs.success - qjobs.has_success
    realrunerror = qjobs.error
    resubjobs = set(
        [k for k, v in qjobs.subtimes.items() if v != qjobs.times])
    thisjobstates = qjobs.state

    if len(realrunerror) == 0:
        print("[%s] All tesks(total(%d), actual(%d), actual_success(%d), actual_error(%d)) in file (%s) finished successfully." %
              (datetime.today().isoformat(), len(thisrunjobs), len(realrunjobs), len(realrunsuccess), len(realrunerror), os.path.abspath(qjobs.jfile)))
    else:
        print "[%s] All tesks( total(%d), actual(%d), actual_success(%d), actual_error(%d) ) in file (%s) finished, But there are ERROR tesks." % (
            datetime.today().isoformat(), len(thisrunjobs), len(realrunjobs), len(realrunsuccess), len(realrunerror), os.path.abspath(qjobs.jfile))
    print dict(Counter(thisjobstates.values()))


def main():
    args = parseArgs()
    global clear
    clear = args.noclean
    h = ParseSingal()
    h.start()
    global qjobs
    qjobs = qsub(args.jobfile, args.num, args.injname,
                 args.start, args.end, mode=args.mode)
    qjobs.run(times=args.resub - 1, resubivs=args.resubivs)
    sumJobs(qjobs)


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# coding:utf-8
'''For summary all jobs

Usage: runstate [jobfile|logdir]
'''

import os
import sys
import psutil
from collections import defaultdict
from commands import getstatusoutput

from job import Jobfile


def style(string, mode='', fore='', back=''):
    STYLE = {
        'fore': {'black': 30, 'red': 31, 'green': 32, 'yellow': 33, 'blue': 34, 'purple': 35, 'cyan': 36, 'white': 37},
        'back': {'black': 40, 'red': 41, 'green': 42, 'yellow': 43, 'blue': 44, 'purple': 45, 'cyan': 46, 'white': 47},
        'mode': {'mormal': 0, 'bold': 1, 'underline': 4, 'blink': 5, 'invert': 7, 'hide': 8},
        'default': {'end': 0},
    }
    mode = '%s' % STYLE["mode"].get(mode, "")
    fore = '%s' % STYLE['fore'].get(fore, "")
    back = '%s' % STYLE['back'].get(back, "")
    style = ';'.join([s for s in [mode, fore, back] if s])
    style = '\033[%sm' % style if style else ''
    end = '\033[%sm' % STYLE['default']['end'] if style else ''
    return '%s%s%s' % (style, string, end)


def main():
    if len(sys.argv) > 2 or "-h" in sys.argv or "--help" in sys.argv:
        print __doc__
        sys.exit(1)
    has_qstat = True if getstatusoutput('command -v qstat')[0] == 0 else False
    username = os.getenv("USER")
    if has_qstat:
        alljobs = os.popen(
            "qstat -u \* | grep -P '^[\d\s]'").readlines()
        jobs = {}
        for j in alljobs:
            j = j.split()
            jobs.setdefault(j[3], defaultdict(int))[j[4]] += 1
        print style("-"*47, mode="bold")
        print style("{0:<20} {1:>8} {2:>8} {3:>8}".format(
            "user", "jobs", "run", "queue"), mode="bold")
        print style("-"*47, mode="bold")
        for u in sorted(jobs.items(), key=lambda x: sum(x[1].values()), reverse=True):
            user = u[0]
            job = sum(u[1].values())
            run = u[1]["r"]
            qw = u[1]["qw"]
            if user == username:
                print style("{0:<20} {1:>8} {2:>8} {3:>8}".format(
                    user, job, run, qw), fore="red", mode="bold")
            else:
                print style("{0:<20} {1:>8} {2:>8} {3:>8}".format(
                    user, job, run, qw))
        print style("-"*47, mode="bold")
        if len(sys.argv) == 2:
            jobfile = sys.argv[1]
            if os.path.isfile(jobfile):

                jf = Jobfile(jobfile)
                jobs = jf.jobs()
                norun = []
                for jn in jf.alljobnames:
                    if not os.path.isfile(os.path.join(jf.logdir, jn + ".log")):
                        norun.append(jn)
                logdir = os.path.abspath(os.path.join(
                    os.path.dirname(jobfile), jf.logdir))
            elif os.path.isdir(jobfile):
                logdir = os.path.abspath(jobfile)
            else:
                raise IOError("No such file or directory %s." % jobfile)
            if not os.path.isdir(logdir):
                raise IOError("No such log_dir %s" % logdir)
            fs = [os.path.join(logdir, i) for i in os.listdir(logdir)]
            # submit = int(os.popen("awk 'FNR==2' " + " ".join(fs) + " | wc -l ").read().strip())
            submit = 0
            stat = []
            for i in fs:
                if not i.endswith("log"):
                    continue
                submit += 1
                s = os.popen("sed -n '$p' " + i).read().strip()
                if s:
                    stat.append(s.split()[-1])
                else:
                    stat.append("")
            # stat = [os.popen("sed -n '$p' " + i).read().strip().split()[-1] for i in fs]
            # stat = [int(os.popen("grep -i ERROR %s | wc -l"%i).read().strip()) for i in fs]
            success = filter(lambda x: x == "SUCCESS", stat)
            error = filter(lambda x: x == "ERROR", stat)
            # running = jobs[username]["r"] + jobs[username]["qw"] if username in jobs else 0
            running = submit - len(success) - len(error)
            print style("-"*47, mode="bold")
            print style("{0:<20} {1:>5}".format("submitted:", submit))
            print style("{0:<20} {1:>5}".format(
                "still runing/queue:", running))
            print style("{0:<20} {1:>5}".format("success:", len(success)))
            print style("{0:<20} {1:>5}".format(
                "error:", len(error)), mode="bold", fore="red")
            print style("-"*47, mode="bold")
            # if len(success) + len(error) + running < submit:
            #    print style("{0} {1}".format("Warning:","some tasks may submite, but not running!" ), mode="bold", fore="red")
            if os.path.isfile(jobfile):
                print "Not submitted jobs: %s" % ", ".join(norun)
    else:
        try:
            users = os.popen(
                "awk -F : '{if($3 > 499){print $1}}' /etc/passwd").read().split()
        except:
            users = map(lambda x: os.path.basename(
                x), os.popen('ls -d /home/*').read().split())
        allpids = psutil.pids()
        jobs = {}
        mem = {}
        for p in allpids:
            try:
                j = psutil.Process(p)
                if j.username() not in users:
                    continue
                RES = j.memory_info().rss
                VIRT = j.memory_info().vms
                jobs.setdefault(j.username(), defaultdict(int))[
                    j.status()] += 1
                mem.setdefault(j.username(), defaultdict(int))["rss"] += RES
                mem.setdefault(j.username(), defaultdict(int))["vms"] += VIRT
            except:
                continue
        print style("-"*70, mode="bold")
        print style("{0:<20} {1:>7} {2:>5} {3:>5} {4:>5} {5:>5} {6:>8} {7:>8}".format(
            "user", "process", "R", "S", "D", "Z", "RES(G)", "VIRT(G)"), mode="bold")
        print style("-"*70, mode="bold")
        for u in sorted(jobs.items(), key=lambda x: sum(x[1].values()), reverse=True):
            user = u[0]
            job = sum(u[1].values())
            res = int(
                round(int(mem[user].get("rss", 0)/1024.0/1024.0/1024.0), 4))
            vms = int(
                round(int(mem[user].get("vms", 0)/1024.0/1024.0/1024.0), 4))
            if user == username:
                print style("{0:<20} {1:>7} {2:>5} {3:>5} {4:>5} {5:>5} {6:>8} {7:>8}".format(user, job, u[1].get("running", 0), u[1].get(
                    "sleeping", 0), u[1].get("disk-sleep", 0), u[1].get("zombie", 0), res, vms), fore="red", mode="bold")
            else:
                print style("{0:<20} {1:>7}".format(user, job)),
                nums = [u[1].get("running", 0), u[1].get("sleeping", 0), u[1].get(
                    "disk-sleep", 0), u[1].get("zombie", 0)]
                printstr = ""
                for i in nums:
                    if i >= 10:
                        printstr += style("{:>5} ".format(i),
                                          fore="red", mode="bold")
                    else:
                        printstr += style("{:>5} ".format(i))
                printstr += style("{:>8} ".format(res)) + \
                    style("{:>8} ".format(vms))
                print printstr.rstrip()
        print style("-"*70, mode="bold")


if __name__ == "__main__":
    main()



#!/usr/bin/env python2
# coding:utf-8

import os
import time
import random

from subprocess import call, PIPE, Popen
from threading import Thread
from Queue import Queue
from datetime import datetime
from collections import defaultdict

from job import Jobfile

RUNSTAT = " && echo [\`date +'%F %T'\`] SUCCESS || echo [\`date +'%F %T'\`] ERROR"


class qsub(object):
    def __init__(self, jobfile, max_jobs=None, jobnames=None, start=1, end=None, mode=None):
        self.pid = os.getpid()
        self.jfile = jobfile
        self.is_run = False
        self.firstjobnames = set()
        self.state = {}

        jf = Jobfile(self.jfile, mode=mode)
        self.has_sge = jf.has_sge
        self.jobs = jf.jobs(jobnames, start, end)  # all jobs defined by args
        self.totaljobs = jf.totaljobs  # all jobs in job file
        self.totaljobdict = {jf.name: jf for jf in self.totaljobs}

        self.orders = jf.orders()  # total orders in job file
        self.orders_rev = {}   # total orders_rev in job file
        for k, v in self.orders.items():
            for i in v:
                self.orders_rev.setdefault(i, set()).add(k)

        # order job name miss
        order_all = set(self.orders.keys() + self.orders_rev.keys())
        if order_all < jf.alljobnames:
            self.extrajob = jf.alljobnames - order_all
            self.throw("There are jobs not defined in orders")

        self.logdir = jf.logdir
        if not os.path.isdir(self.logdir):
            os.makedirs(self.logdir)

        self.error = set()  # args jobs error
        self.success = set()  # args jobs success,   self.error + self.success = len(self.jobs)
        self.thisjobnames = set([j.name for j in self.jobs])
        self.has_success = set()

        for jn in self.thisjobnames.copy():
            lf = os.path.join(self.logdir, jn + ".log")
            if os.path.isfile(lf):  # if log file not exists, will run
                js = self.jobstatus(jn)
                if js != "success":
                    os.remove(lf)
                    if js == "error":
                        self.error.remove(jn)
                        self.state.pop(jn)
                else:
                    self.thisjobnames.remove(jn)  # thisjobs - successjobs
                    self.has_success.add(jn)
            else:
                self.state[jn] = "wait"
        # thisjobnames are real jobs
        # len(self.has_success) + len(self.thisjobnames) = len(self.jobs)
        self.max_jobs = len(
            self.thisjobnames) if max_jobs is None else max_jobs

        self.jobqueue = Queue(maxsize=self.max_jobs)

    def not_qsub(self, jobname):
        qs = os.popen('qstat -xml | grep %s_%d | wc -l' %
                      (jobname, self.pid)).read().strip()
        if int(qs) == 0:
            return True
        return False

    def jobstatus(self, jobname):
        status = "wait"  # wait to submit
        logfile = os.path.join(self.logdir, jobname + ".log")
        if os.path.isfile(logfile):
            status = "submit"  # wait to run
            try:
                sta = os.popen('tail -n 1 %s' % logfile).read().split()[-1]
            except IndexError:
                status = "run"
                # if self.not_qsub(jobname) and self.is_run:                                                           ## job exit, qsub error
                #    self.throw("Error in %s job, probably because of qsub interruption."%jobname)
                return status
            if sta == "SUCCESS":
                status = "success"
                self.success.add(jobname)
            elif sta == "ERROR":
                status = "error"
                self.error.add(jobname)
            elif sta == "Exiting.":
                status = "exit"
            else:
                if "RUNNING..." in os.popen("sed -n '3p' %s" % logfile).read():
                    status = "run"
                    # if self.not_qsub(jobname) and self.is_run:                                                       ## job exit, qsub error
                    #    self.throw("Error in %s job, probably because of qsub interruption."%jobname)
        self.state[jobname] = status
        return status

    def firstjob(self):
        queryjob_tmp = self.thisjobnames.copy()
        for j in self.thisjobnames:
            if j in self.orders:
                for bj in self.orders[j]:
                    if bj in self.thisjobnames:
                        queryjob_tmp.remove(j)
                        break
        self.firstjobnames.update(queryjob_tmp)
        return [self.totaljobdict[i] for i in self.firstjobnames]

    def qsubCheck(self, num, sec=1, ):
        qs = 0
        while True:
            time.sleep(sec)  # check per 1 seconds if job pools
            if not self.jobqueue.full():
                continue
            # qs = os.popen('qstat -xml | grep _%d | wc -l' %
            #              self.pid).read().strip()
            #qs = int(qs)
            qs = len([i for i, j in self.state.items()
                      if j in ["run", "submit", "resubmit"]])
            if qs < num:
                [self.jobqueue.get() for _ in range(num-qs)]
            else:
                continue

    def run(self, sec=2, times=-1, resubivs=2):

        self.is_run = True
        self.times = times
        self.subtimes = defaultdict(lambda: self.times)

        firstqsub = self.firstjob()
        if self.max_jobs < 100:
            p = Thread(target=self.qsubCheck, args=(self.max_jobs,))
            p.setDaemon(True)
            p.start()
        prepare_sub = set()
        for job in firstqsub:
            self.submit(job)
            if job.name in self.orders_rev:
                # prepare_sub.update(
                #    [i for i in self.orders_rev[job.name] if self.jobstatus(i) != "success"])
                prepare_sub.update(
                    [i for i in self.orders_rev[job.name] if i in self.thisjobnames])
        while len(self.thisjobnames) > 0:
            time.sleep(sec)
            for k in prepare_sub.copy():
                time.sleep(0.1)
                subK = True
                for jn in self.orders[k]:
                    time.sleep(0.1)
                    js = self.jobstatus(jn)
                    if js == "success":
                        continue
                    elif js == "error":
                        if self.subtimes[jn] < 0:
                            continue
                            # self.throw("Error jobs return, %s"%os.path.join(self.logdir, jn + ".log"))   ## if error, exit program
                        else:
                            time.sleep(resubivs)  # sleep, re-submit
                            self.submit(self.totaljobdict[jn], resub=True)
                            self.subtimes[jn] -= 1
                            subK = False
                    elif js == "exit":
                        self.throw("Error when qsub")
                    else:
                        subK = False
                if subK:
                    self.submit(self.totaljobdict[k])
                    if k in prepare_sub:
                        prepare_sub.remove(k)
                    if k in self.orders_rev:
                        prepare_sub.update(self.orders_rev[k])
                    for jn in self.orders[k]:
                        if jn in prepare_sub:
                            prepare_sub.remove(jn)
        self.finalstat(resubivs)

    def submit(self, job, resub=False):
        logfile = os.path.join(self.logdir, job.name + ".log")

        if self.jobstatus(job.name) == "success":
            if job.name in self.thisjobnames:
                self.thisjobnames.remove(job.name)
            return

        if resub:
            logcmd = open(logfile, "a")
            logcmd.write("\n" + job.cmd+"\n")
            self.state[job.name] = "resubmit"
        else:
            logcmd = open(logfile, "w")
            logcmd.write(job.cmd+"\n")
        logcmd.write("[%s] " % datetime.today().strftime("%F %X"))
        logcmd.flush()
        if self.max_jobs < 100:
            self.jobqueue.put(job.name, block=True, timeout=1080000)
        qsubline = "echo [\`date +'%F %T'\`] RUNNING... && " + \
            job.cmd + RUNSTAT
        if job.host is not None and job.host == "localhost":
            cmd = '''echo "Your job \\('%s'\\) has been submitted in localhost" && ''' % job.name + qsubline
            cmd = cmd.replace("\\", "")
            if resub:
                cmd = cmd.replace("RUNNING", "RUNNING \\(re-run\\)")
            Popen(cmd, shell=True, stdout=logcmd, stderr=logcmd)
        else:
            job.cmd = job.cmd.replace('"', "\\\"")
            cmd = 'qsub %s -N %s_%d -o %s -j y <<< "%s"' % (
                job.sched_options, job.name, self.pid, logfile, qsubline)
            if resub:
                cmd = cmd.replace("RUNNING", "RUNNING \\(re-submit\\)")
            call(cmd, shell=True, stdout=logcmd, stderr=logcmd)
        logcmd.close()
        if job.name in self.thisjobnames:
            self.thisjobnames.remove(job.name)

    def finalstat(self, resubivs):
        finaljobs = set([j.name for j in self.jobs]) - \
            self.success - self.error
        while len(finaljobs) > 0:
            time.sleep(2)
            for jn in finaljobs.copy():
                js = self.jobstatus(jn)
                if js == "success":
                    finaljobs.remove(jn)
                elif js == "error":
                    if self.subtimes[jn] < 0:
                        finaljobs.remove(jn)
                    else:
                        time.sleep(resubivs)  # sleep, re-submit
                        self.submit(self.totaljobdict[jn], resub=True)
                        self.subtimes[jn] -= 1

    def throw(self, msg):
        raise RuntimeError(msg)



import os

from setuptools import setup
from src.version import __version__


def getdes():
    des = ""
    with open(os.path.join(os.getcwd(), "README.md")) as fi:
        des = fi.read()
    return des


setup(
    name="runjob",
    version=__version__,
    packages=["runjob"],
    package_dir={"runjob": "src"},
    data_files=[("doc", ["doc/jobA.log", "doc/jobB.log", "doc/example.job"]), ],
    author="Deng Yong",
    author_email="yodeng@tju.edu.cn",
    url="https://github.com/yodeng/runjob",
    license="BSD",
    install_requires=["psutil"],
    classifiers=[
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
    ],
    long_description=getdes(),
    entry_points={
        'console_scripts': [
            'runjob = runjob.main:main',
            'runstate = runjob.stat:main',
            'runsge = runjob.sge:main',
        ]
    }
)
